                             
' /////////////////////////////////////////////////////////////////////////////
' Donkey Kong Remake
' AUTHOR: Steve Waddicor
'//////////////////////////////////////////////////////////////////////////////

CON
    _clkmode = xtal1 + pll8x
    _xinfreq = 10_000_000 + 0000
    _stack = 100 'accomodate display memory and stack
    cogs = 5

CON
    SPRITE_FLAGS                =3
    SPRITE_TILE                 =2
    SPRITE_XPOS                 =1
    SPRITE_YPOS                 =0 
    SPRITE_INVISIBLE            =%100
    SPRITE_FLIPPED              =%010
    SPRITE_MIRRORED             =%001
    
VAR
    ' Renderer parameter block
    ' Initializing data is in the DAT block at the end of this file.
    ' This is copied in to this variable space with a longmove
    ' so it can be altered easily (without remembering to use @ all the time).
    ' Each renderer then copies these values to it's own cog_memory cache
    ' when it's got a little spare time, several times per frame
    ' so these values can be altered on the fly. 
    long  rend_scanline_req_adr    'a pointer to the request to put a line in the scan buffer   write-only
    long  rend_map                 'pointer to screen (bytes)                                write-only
    long  rend_tiletable           'pointer to tile definitions (words)                      write-only
    long  rend_spritetable         'pointer to tile definitions (longs)                      write-only
    long  rend_colors              'pointer to colors (longs)                                write-only               
    long  rend_ht                  'horizontal tiles                                         write-only
    long  rend_vt                  'vertical tiles                                           write-only
    long  rend_scanline_buffer_adr 'hug memory where renderer puts scanline when requested   write-only
    long  rend_cog_number          'each cog gets a different number, starting with zero     write-only
    long  rend_cog_count           'total number of cogs                                     write-only
    long  rend_sprites[32]         '%_____ifm_nnnnnnnn_xxxxxxxx_yyyyyyyy    (invisible, flipped, mirrored)
    long  rend_debug_map_ptr
    'End of redderer parameter block

VAR
    long  startkey
    long  debug
    byte  debug_map[23*28]

OBJ

  tv    : "sw_dk_tv_drv_019.spin"
  gfx   : "sw_dkong_gfx_003.spin"
  rend  : "sw_dk_gfx_renderer_009.spin"

CON
    NUMBER_OF_BARRELS

VAR
    long    gp
    long    mario_state
    long    mario_state_count
    long    game_state
    long    kong_state
    long    kong_state_count
    long    barrel_state[10]
    long    barrel_state_count[10]
    long    kong_barrel_releases
    long    kong_holding_barrel
    long    frame
    
CON
    'states where user input accepted
    MARIO_FACING_LEFT           = 0
    MARIO_FACING_RIGHT          = 1
    MARIO_WALK_LEFT             = 2
    MARIO_WALK_RIGHT            = 3
    MARIO_HAMMER_LEFT           = 4
    MARIO_HAMMER_RIGHT          = 5
    MARIO_CLIMB_UP              = 6
    MARIO_CLIMB_DOWN            = 7
    'no user input accepted
    MARIO_JUMP_LEFT             = 8
    MARIO_JUMP_RIGHT            = 9
    MARIO_JUMP_UP_FACING_LEFT   = 10    
    MARIO_JUMP_UP_FACING_RIGHT  = 11
    MARIO_LANDING_LEFT          = 12
    MARIO_LANDING_RIGHT         = 13

CON
    GAME_LEVEL_1                = 0
    GAME_LEVEL_2                = 1
    GAME_LEVEL_3                = 2
    GAME_LEVEL_4                = 3
    GAME_INIT_LEVEL_1           = 4
    GAME_INIT_LEVEL_2           = 5
    GAME_INIT_LEVEL_3           = 6
    GAME_INIT_LEVEL_4           = 7    

CON
    KONG_LEVEL_START_POSE       = 0
    KONG_LEFT                   = 1
    KONG_CENTER_1               = 2
    KONG_RIGHT                  = 3
    KONG_CENTER_2               = 4
    KONG_GROWL                  = 5
    KONG_LEVEL_START_POSE_INIT  = 11
    KONG_LEFT_INIT              = 6
    KONG_CENTER_1_INIT          = 7
    KONG_RIGHT_INIT             = 8
    KONG_CENTER_2_INIT          = 9
    KONG_GROWL_INIT             = 10

CON
    BARREL_STILL                = 0
    BARREL_ROLLING_RIGHT        = 1
    BARREL_ROLLING_LEFT         = 2
    BARREL_FALLING_RIGHT        = 3
    BARREL_FALLING_LEFT         = 4
    BARREL_BOUNCING_RIGHT       = 5
    BARREL_BOUNCING_LEFT        = 6
    BARREL_ON_LADDER            = 7
    BARREL_FALLING_DOWN         = 8
    BARREL_FALLING_ACROSS       = 9
     
DAT

kong_center_sad_template
                        long   $00_22_34_24            'torso
                        long    $00_24_40_24            'right leg
                        long    $01_24_28_24            'left leg
                        long    $01_20_34_14            'head
                        long    $00_26_2a_18            'left arm
                        long    $01_26_3e_18            'right arm
                        long    $04_00_00_00            'empty
                        long    $04_00_00_00            'empty

kong_center_roar_template
                        long $00_22_34_24             'torso
                        long    $00_24_40_24            'right leg
                        long    $00_23_28_24            'left leg
                        long    $00_25_40_14            'right arm
                        long    $01_21_34_14            'head
                        long    $00_26_2a_18            'left arm
                        long    $04_00_00_00            'empty
                        long    $04_00_00_00            'empty

kong_center_barrel_template
                        long $00_22_34_24           'torso
                        long    $00_24_40_24            'right leg
                        long    $01_24_28_24            'left leg
                        long    $01_20_34_14            'head
                        long    $01_28_24_22            'left arm
                        long    $00_28_44_22            'right arm
                        long    $01_27_24_12            'left shoulder
                        long    $00_27_44_12            'right ahoulder

' middle barrel         long    $00_16_34_24            'empty

kong_left_template
                        long $00_2b_32_24            'torso
                        long    $00_29_42_24            'right leg
                        long    $00_2a_22_24            'left leg
                        long    $00_2c_32_14            'shoulder
                        long    $00_2d_22_14            'head
                        long    $04_00_00_00            'empty
                        long    $04_00_00_00            'empty
                        long    $04_00_00_00            'empty

' barrel behind stack   long    $00_14_16_26            'empty

                        

kong_right_template
                        long    $01_2b_34_24            'torso
                        long    $01_29_24_24            'right leg
                        long    $01_2a_44_24            'left leg
                        long    $01_2c_34_14            'shoulder
                        long    $01_2d_44_14            'head
                        long    $04_00_00_00            'empty
                        long    $04_00_00_00            'empty
                        long    $04_00_00_00            'empty

                            
PUB Main | i
    'Start renderers
    rend_tiletable := gfx.tile_adr
    rend_spritetable := gfx.sprite_adr
    rend_map := gfx.map_adr
    rend_ht := 32
    rend_vt := 28
    rend_scanline_buffer_adr := @hub_scanline_buffer
    rend_scanline_req_adr := @scanline_req
    rend_cog_count := cogs
    repeat i from 0 to 31
        rend_sprites[i] := $04_00_00_00                 'make them all invisible by default


    rend_debug_map_ptr := 0
    repeat i from 0 to constant(cogs-1)
        rend_cog_number := i
        rend.start(@rend_scanline_req_adr)
        'Wait to allow this cog to get the cog number param from the block before changing it.
        repeat 1000
     
    ' start the TV driver 
    tv.start(@tv_params)
    
    DebugInit(@debug_map)
    rend_debug_map_ptr:=0
    

    game_state:=GAME_INIT_LEVEL_1

    'Main Game Loop
    repeat
        DebugInit(@debug_map)
        gp := NES_Read_Gamepad

        case game_state
            GAME_INIT_LEVEL_1:
                InitLevel1
            GAME_LEVEL_1..GAME_LEVEL_4:    
               ' if !(rend_sprites.byte[SPRITE_FLAGS]&SPRITE_INVISIBLE)
                    MarioLogic
                KongLogic
                BarrelLogic
 
        if gp & %00010000
            if startkey==0
                startkey:=1
                !debug
                if debug
                    rend_debug_map_ptr:=@debug_map
                else
                    rend_debug_map_ptr:=0
        else
            startkey:=0
        
        'repeat 500

CON
    NUMBER_OF_MOVING_BARRELS    = 10
    'sprite indexes    
    MARIO_SPRITE                = 0
    KONG_SPRITES                = MARIO_SPRITE+1
    MOVING_BARREL_SPRITES       = KONG_SPRITES+8
    FIXED_BARREL_SPRITES        = MOVING_BARREL_SPRITES+NUMBER_OF_MOVING_BARRELS
    OIL_BARREL                  = FIXED_BARREL_SPRITES+4
    ' Kong sends a blue barrel every 8th time
    ' Blue ones go all the way to the bottom.  Ordinary ones sneak off the sides when past.
        
PRI InitLevel1 | i
    rend_sprites[MARIO_SPRITE]           := $00_00_34_c8
    rend_sprites[FIXED_BARREL_SPRITES ]  := $00_17_0c_13
    rend_sprites[FIXED_BARREL_SPRITES+1] := $00_17_0c_23
    rend_sprites[FIXED_BARREL_SPRITES+2] := $00_17_16_13
    rend_sprites[FIXED_BARREL_SPRITES+3] := $00_17_16_23
    rend_sprites[OIL_BARREL] := $00_42_20_c8
    repeat i from 0 to constant(NUMBER_OF_BARRELS-1)
        barrel_state[i] := BARREL_STILL
    game_state := GAME_LEVEL_1
    kong_state := KONG_LEVEL_START_POSE_INIT
    frame := SPRITE_INVISIBLE<<8
    
PRI MarioLogic | current_height, platform_height, ladder_top, ladder_bottom, delta
    current_height := rend_sprites.byte[SPRITE_YPOS]+15
    platform_height := PlatformHeight(rend_sprites.byte[SPRITE_XPOS]+8,current_height) 
    ladder_top := LadderTopHeight(rend_sprites.byte[SPRITE_XPOS]+8,current_height)
    ladder_bottom := LadderBottomHeight(rend_sprites.byte[SPRITE_XPOS]+8,current_height)
    debugdec(@currentheighttext,current_height)
    debugdec(@laddertoptext,ladder_top)
    debugdec(@ladderbottomtext,ladder_bottom)
    debugdec(@platformheighttext,platform_height)
     
    ' decide on next state, and do any unique initialisations for that state
    if (mario_state=<MARIO_CLIMB_DOWN)
        'but only if we have user control
        if (gp & GP_A) and (mario_state==MARIO_WALK_LEFT)
            mario_state:=MARIO_JUMP_LEFT
            frame := $00d
            mario_state_count:=0
            
        elseif (gp & GP_A) and (mario_state==MARIO_WALK_RIGHT)
            mario_state:=MARIO_JUMP_RIGHT
            frame := $10d
            mario_state_count:=0
         
        elseif (gp & GP_A) and (mario_state==MARIO_FACING_LEFT)
            mario_state:=MARIO_JUMP_UP_FACING_LEFT
            frame := $00d
            mario_state_count:=0
         
        elseif (gp & GP_A) and (mario_state==MARIO_FACING_RIGHT)
            mario_state:=MARIO_JUMP_UP_FACING_RIGHT
            frame := $10d
            mario_state_count:=0
         
        elseif (gp & GP_LEFT) and (current_height==platform_height)
            rotate(@mario_state_count,1,0,12)
            if mario_state_count//3 <> 1
                rend_sprites.byte[SPRITE_XPOS]-=1
            frame := lookupz(mario_state_count: 0,0,0,2,2,2,0,0,0,1,1,1)              
            mario_state:=MARIO_WALK_LEFT
         
        elseif (gp & GP_RIGHT) and (current_height==platform_height)
            rotate(@mario_state_count,1,0,12)
            if mario_state_count//3 <> 2
                rend_sprites.byte[SPRITE_XPOS]+=1
            frame := lookupz(mario_state_count: 0,0,0,2,2,2,0,0,0,1,1,1) | SPRITE_MIRRORED<<8
            mario_state:=MARIO_WALK_RIGHT
         
        elseif (gp & GP_UP) and (ladder_top<current_height)
            if (mario_state<>MARIO_CLIMB_UP) and (mario_state<>MARIO_CLIMB_DOWN)
                mario_state_count:=0
            else
                mario_state_count+=1
            if mario_state_count//5==0
                rend_sprites.BYTE[SPRITE_YPOS]-=2
                rend_sprites.BYTE[SPRITE_YPOS]<#=platform_height-15
            mario_state:=MARIO_CLIMB_UP
                 
        elseif (gp & GP_DOWN) and (ladder_bottom>current_height)
            if (mario_state<>MARIO_CLIMB_UP) and (mario_state<>MARIO_CLIMB_DOWN)
                mario_state_count:=0
            else
                mario_state_count-=1
            if mario_state_count//5==0
                rend_sprites.BYTE[SPRITE_YPOS]+=2
            mario_state:=MARIO_CLIMB_DOWN
         
        elseif (mario_state==MARIO_WALK_LEFT)
            mario_state:=MARIO_FACING_LEFT
         
        elseif (mario_state==MARIO_WALK_RIGHT)
            mario_state:=MARIO_FACING_RIGHT
    DebugDec(@errortext,mario_state)
         
    'do the things that are common between states and that are done every game loop for a state
    if (mario_state==MARIO_CLIMB_DOWN) and (current_height==ladder_bottom-1)
        ' catch mario just before he goes below platform if he's on odd height.
        rend_sprites.byte[SPRITE_YPOS]:=ladder_bottom-15
        current_height:=ladder_bottom
    case mario_state
        MARIO_CLIMB_UP,MARIO_CLIMB_DOWN :
            case(ladder_bottom-current_height)/2
                0..1 : frame := 6
                other:
                    delta := (current_height-ladder_top)/2
                    case delta
                        0 : frame := 6
                        1 : frame := 3
                        2..3 : frame := 5
                        4..5 : frame := $104
                        other : frame := (delta//2)*$100+3
            rend_sprites.byte[SPRITE_XPOS]&=$F8
            rend_sprites.byte[SPRITE_XPOS]+=4
            if current_height==platform_height
                mario_state:=MARIO_FACING_LEFT                              'but, frame is facing away
     
        MARIO_WALK_LEFT, MARIO_WALK_RIGHT:
            'deal with walking on platforms, inclines and falling   
            rend_sprites.byte[SPRITE_YPOS] := PlatformHeight(rend_sprites.byte[SPRITE_XPOS]+8,rend_sprites.byte[SPRITE_YPOS]+15)-15
             
        MARIO_JUMP_LEFT:
            'we lookup the y delta for each frame.  By offsetting by one, we ensure that overflow means one will be subtracted.
            rend_sprites.BYTE[SPRITE_YPOS]-=lookupz(mario_state_count: 3,2,2,2,2,2,2,2,1,2,2,1,2,1,2,1,1,1,2,1,1,1,0,1,1,1,0,1,0,1,0,0,1)-1
            if mario_state_count//2==0
                rend_sprites.BYTE[SPRITE_XPOS]-=1                 
            mario_state_count+=1
     
        MARIO_JUMP_RIGHT:
            'we lookup the y delta for each frame.  By offsetting by one, we ensure that overflow means one will be subtracted.
            rend_sprites.BYTE[SPRITE_YPOS]-=lookupz(mario_state_count: 3,2,2,2,2,2,2,2,1,2,2,1,2,1,2,1,1,1,2,1,1,1,0,1,1,1,0,1,0,1,0,0,1)-1
            if mario_state_count//2==0
                rend_sprites.BYTE[SPRITE_XPOS]+=1                 
            mario_state_count+=1
     
        MARIO_JUMP_UP_FACING_LEFT:
            'we lookup the y delta for each frame.  By offsetting by one, we ensure that overflow means one will be subtracted.
            rend_sprites.BYTE[SPRITE_YPOS]-=lookupz(mario_state_count: 3,2,2,2,2,2,2,2,1,2,2,1,2,1,2,1,1,1,2,1,1,1,0,1,1,1,0,1,0,1,0,0,1)-1
            mario_state_count+=1
     
        MARIO_JUMP_UP_FACING_RIGHT:
            'we lookup the y delta for each frame.  By offsetting by one, we ensure that overflow means one will be subtracted.
            rend_sprites.BYTE[SPRITE_YPOS]-=lookupz(mario_state_count: 3,2,2,2,2,2,2,2,1,2,2,1,2,1,2,1,1,1,2,1,1,1,0,1,1,1,0,1,0,1,0,0,1)-1
            mario_state_count+=1
     
        MARIO_LANDING_LEFT:
            if mario_state_count==4
                mario_state:=MARIO_FACING_LEFT
                frame := $000
            else
                mario_state_count++
     
        MARIO_LANDING_RIGHT:
            if mario_state_count==4
                mario_state:=MARIO_FACING_RIGHT
                frame := $100
            else
                mario_state_count++
     
    'catch jump states that have past completion 
    case mario_state
        MARIO_JUMP_LEFT,MARIO_JUMP_UP_FACING_LEFT:
            if current_height>platform_height and mario_state_count>32 'At least 32 frames into the jump to avoid latching on platforms above
                mario_state:=MARIO_LANDING_LEFT
                frame:=$00e
                mario_state_count:=0
                rend_sprites.BYTE[SPRITE_YPOS]:=platform_height-15
            elseif rend_sprites.byte[SPRITE_XPOS]<16
                mario_state:=MARIO_JUMP_RIGHT
                frame := $10d
                
        MARIO_JUMP_RIGHT,MARIO_JUMP_UP_FACING_RIGHT:
            if current_height>platform_height and mario_state_count>32 'At least 32 frames into the jump to avoid latching on platforms above 
                mario_state:=MARIO_LANDING_RIGHT
                frame:=$10e
                mario_state_count:=0
                rend_sprites.BYTE[SPRITE_YPOS]:=platform_height-15
            elseif rend_sprites.byte[SPRITE_XPOS]>226
                mario_state:=MARIO_JUMP_LEFT
                frame := $00d
     
    rend_sprites.byte[SPRITE_XPOS]#>=16        
    rend_sprites.byte[SPRITE_XPOS]<#=226
    rend_sprites.WORD[1]:=frame

PRI KongLogic
    debugdec(@kongstatetext,kong_state)
    kong_state_count++
    case kong_state
        KONG_LEVEL_START_POSE_INIT:
            kong_barrel_releases:=0
            kong_state_count := 0 
            longmove(@rend_sprites[kong_sprites], @kong_center_sad_template, 8)
            kong_state := KONG_LEVEL_START_POSE
        
        KONG_LEFT_INIT:
            kong_holding_barrel := kong_barrel_releases//10
            barrel_state[kong_holding_barrel] := BARREL_STILL
            if kong_barrel_releases&$7 == 0
                rend_sprites[MOVING_BARREL_SPRITES+kong_holding_barrel] := $00_18_16_26
            else
                rend_sprites[MOVING_BARREL_SPRITES+kong_holding_barrel] := $00_14_16_26
            kong_state_count := 0
            longmove(@rend_sprites[kong_sprites], @kong_left_template, 8)
            kong_state := KONG_LEFT
        
        KONG_CENTER_1_INIT:
            ' keep barrel type
            debugdec(@knongbarrelreleasestext,kong_barrel_releases)
            'repeat 1000000
            rend_sprites[MOVING_BARREL_SPRITES+kong_holding_barrel] += $00_01_1E_00
            if kong_barrel_releases==0
                barrel_state[kong_holding_barrel] := BARREL_FALLING_DOWN
                kong_barrel_releases++                 
                kong_state := KONG_CENTER_2_INIT
            else
                kong_state := KONG_CENTER_1
            kong_state_count := 0
            longmove(@rend_sprites[kong_sprites], @kong_center_barrel_template, 8)
  
        KONG_RIGHT_INIT:
            ' keey barrel type
            rend_sprites[MOVING_BARREL_SPRITES+kong_holding_barrel] -= $00_01_1E_00
            ' and add on x delta
            rend_sprites[MOVING_BARREL_SPRITES+kong_holding_barrel] += $00_00_3C_00
            kong_state_count := 0
            longmove(@rend_sprites[kong_sprites], @kong_right_template, 8)
            kong_state := KONG_RIGHT

        KONG_CENTER_2_INIT:
            debugdec(@kongbarreltext, kong_holding_barrel)
            kong_state_count := 0
            longmove(@rend_sprites[kong_sprites], @kong_center_sad_template, 8)
            kong_state := KONG_CENTER_2

        KONG_GROWL_INIT:
            kong_state_count := 0
            longmove(@rend_sprites[kong_sprites], @kong_center_roar_template, 8)
            kong_state := KONG_GROWL

        KONG_LEVEL_START_POSE:
            if kong_state_count == 64
                mario_state:=MARIO_WALK_RIGHT
                mario_state_count := 0
                frame := SPRITE_MIRRORED<<8
                kong_state := KONG_LEFT_INIT

        KONG_LEFT:
            debugdec(@kongstatecounttext,kong_state_count)
            if kong_state_count == 24
                kong_state := KONG_CENTER_1_INIT

        KONG_CENTER_1:
            if kong_state_count == 24
                kong_state := KONG_RIGHT_INIT
                    
        KONG_RIGHT:
            if kong_state_count == 24
                kong_state := KONG_CENTER_2_INIT
                barrel_state[kong_holding_barrel]:=BARREL_ROLLING_RIGHT
                kong_barrel_releases++

        KONG_CENTER_2:
            if kong_barrel_releases//19 == 18
                kong_state := KONG_GROWL_INIT
            elseif kong_state_count == 24
                kong_state := KONG_LEFT_INIT

        KONG_GROWL:
            if kong_state_count == 53
                kong_state := KONG_LEFT_INIT

PRI BarrelLogic | i, this_barrel, platform_height, height_to_fall
    repeat i from 0 to constant(NUMBER_OF_MOVING_BARRELS-1)
        debugdec(@barrelstatetext,barrel_state[i])
        this_barrel := @rend_sprites[MOVING_BARREL_SPRITES+i]
        if byte[this_barrel][SPRITE_XPOS]==$24 and byte[this_barrel][SPRITE_YPOS]>200
            barrel_state[i]:=BARREL_STILL
            byte[this_barrel][SPRITE_FLAGS]:=SPRITE_INVISIBLE    
        case barrel_state[i]
            BARREL_ROLLING_RIGHT:
                platform_height := PlatformHeight(byte[this_barrel][SPRITE_XPOS]+8,byte[this_barrel][SPRITE_YPOS]+12)
                if platform_height-(byte[this_barrel][SPRITE_YPOS]+12) > 3
                    barrel_state_count[i]:=0
                    barrel_state[i] := BARREL_FALLING_RIGHT
                else 
                    byte[this_barrel][SPRITE_XPOS]+=1
                    byte[this_barrel][SPRITE_YPOS]:=platform_height-12
            BARREL_ROLLING_LEFT:
                platform_height := PlatformHeight(byte[this_barrel][SPRITE_XPOS]+8,byte[this_barrel][SPRITE_YPOS]+12)
                if platform_height-(byte[this_barrel][SPRITE_YPOS]+12) > 3
                    barrel_state_count[i]:=0
                    barrel_state[i] := BARREL_FALLING_LEFT
                else 
                    byte[this_barrel][SPRITE_XPOS]-=1
                    byte[this_barrel][SPRITE_YPOS]:=platform_height-12
            BARREL_FALLING_RIGHT:
                platform_height := PlatformHeight(byte[this_barrel][SPRITE_XPOS]+8,byte[this_barrel][SPRITE_YPOS]+12)
                height_to_fall := platform_height-(byte[this_barrel][SPRITE_YPOS]+12)
                debugdec(@heighttofalltext,height_to_fall)
                if height_to_fall < 0
                    barrel_state_count[i]:=0
                    barrel_state[i] := BARREL_BOUNCING_LEFT    
                else
                    ' offset by 2 so that any additional falls will be 2
                    byte[this_barrel][SPRITE_YPOS]+=lookupz(barrel_state_count[i]: {-1,}-2,-2,-2,-2,-2,-2,-2,-2,-2,-1,-2,-2,-1,-1,-2,-1,-1,-1,-1,-2,0,-1,-1,-1,-1,0,-1,-1,0,-1)+2
                    byte[this_barrel][SPRITE_XPOS]+=lookupz(barrel_state_count[i]: { 1,} 1, 1, 1, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0,0, 1, 0, 1, 0,0, 1, 0,0, 1)
                    barrel_state_count[i]++
            BARREL_FALLING_LEFT:
                platform_height := PlatformHeight(byte[this_barrel][SPRITE_XPOS]+8,byte[this_barrel][SPRITE_YPOS]+12)
                height_to_fall := platform_height-(byte[this_barrel][SPRITE_YPOS]+12)
                debugdec(@heighttofalltext,height_to_fall)
                if height_to_fall < 0
                    barrel_state_count[i]:=0
                    barrel_state[i] := BARREL_BOUNCING_RIGHT    
                else
                    ' offset by 2 so that any additional falls will be 2
                    byte[this_barrel][SPRITE_YPOS]+=lookupz(barrel_state_count[i]: {-1,}-2,-2,-2,-2,-2,-2,-2,-2,-2,-1,-2,-2,-1,-1,-2,-1,-1,-1,-1,-2,0,-1,-1,-1,-1,0,-1,-1,0,-1)+2
                    byte[this_barrel][SPRITE_XPOS]-=lookupz(barrel_state_count[i]: { 1,} 1, 1, 1, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0,0, 1, 0, 1, 0,0, 1, 0,0, 1)
                    barrel_state_count[i]++
            BARREL_BOUNCING_RIGHT:
                if barrel_state_count[i] =< 22
                    byte[this_barrel][SPRITE_YPOS]-=lookupz(barrel_state_count[i]: 1,0,0,1,0,0,0,0,0,-1,0,0,-1,0,0,-1,0,1,0,0,0,0,-1)
                    barrel_state_count[i]++                 
                    byte[this_barrel][SPRITE_XPOS]+=1
                else
                    barrel_state[i]:=BARREL_ROLLING_RIGHT
            BARREL_BOUNCING_LEFT:
                if barrel_state_count[i] =< 22
                    byte[this_barrel][SPRITE_YPOS]-=lookupz(barrel_state_count[i]: 1,0,0,1,0,0,0,0,0,-1,0,0,-1,0,0,-1,0,1,0,0,0,0,-1)
                    barrel_state_count[i]++                 
                    byte[this_barrel][SPRITE_XPOS]-=1
                else
                    barrel_state[i]:=BARREL_ROLLING_LEFT
            BARREL_ON_LADDER:
            BARREL_FALLING_DOWN:
                platform_height := PlatformHeight(byte[this_barrel][SPRITE_XPOS]+8,byte[this_barrel][SPRITE_YPOS]+12)
                height_to_fall := platform_height-(byte[this_barrel][SPRITE_YPOS]+12)
                debugdec(@heighttofalltext,height_to_fall)
                if height_to_fall>0 and height_to_fall<2 and barrel_state_count[i]>10
                    barrel_state_count[i]:=0
                    if platform_height>200
                        byte[this_barrel][SPRITE_YPOS]-=1
                        byte[this_barrel][SPRITE_TILE] := $18
                        barrel_state[i]:=BARREL_BOUNCING_LEFT
                else
                    ' offset by 2 so that any additional falls will be 2
                    byte[this_barrel][SPRITE_YPOS]+=lookupz(barrel_state_count[i]: -2,-2,-2,-2,-2,-2,-1,-2,-1,-2,-1,-2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,0,-1,0,-1,0,-1)+2
                    barrel_state_count[i]++
                
            'BARREL_ROLLING_RIGHT_INIT: 
            'BARREL_ROLLING_LEFT_INIT: 
            'BARREL_FALLING_RIGHT_INIT:
            'BARREL_FALLING_LEFT_INIT:
            'BARREL_EXITING_RIGHT_INIT:
            'BARREL_EXITING_LEFT_INIT:
            'BARREL_ON_LADDER_INIT:

PRI Rotate(item_ptr,increment,lower,upper)
    ' Rotate a byte value through a range of values.  Useful for sprite animation.
    ' byte_item = pointer to a variable to ,reset,lower,upper)
    ' increment = value to add or subtract
    ' lower     = lower bound
    ' upper     = upper bound
    ' if the value is out of range, this doesn't try to apply limits if moving towards the loop.
    long[item_ptr]+=increment
    if increment>0
        if long[item_ptr]>upper
            long[item_ptr] := lower
        elseif long[item_ptr]<lower
            long[item_ptr] := upper
    return

PRI TileAt(x,y) | pos
    pos := (x)>>3+((y)>>3)*32
    result := byte[rend_map][pos]
    

PRI TileHeight(tile)    
    if (tile=>143 and tile=<144)
        return tile-143
    if (tile=>159 and tile=<166)
        return tile-159
    return 0

PRI PlatformHeight(_x,_y) | background
    ' Calculates the position of one pixel above height of the platform where we are.
    ' If we are stood on the platform it's where our feet are.
    ' If our feet are overlapping a platform it's the top of that.(To cope with inclines.)
    ' Otherwise it's the first platform below.           
    repeat
        background := TileAt(_x,_y)
        case background
            135..142:       _y+=8                                        'tile with ladder below
            143..150:       return (_y & $F8)+(background-144)      
            151:            _y+=8                                        'empty space
            152..158:       _y-=8                                        'platform with space below
            159..168:       return (_y & $F8)+(background-160)
            other:          return (_y & $F8)

PRI LadderTopHeight(_x,_y) | background
    ' Calculates the position of the platform at the top of a ladder.  
    ' If we are by a ladder, then this one pixel above the platform at the top of the ladder,
    ' Otherwise it's the same value as LowPlatformHeight( ).
    ' If it's a broken ladder, then it's still the position of the platform above.
    repeat
        background := TileAt(_x,_y)
        case background
            135..150,178:   _y-=8                                         'ladders + workaround invisible ladder
            159..166:       return (_y & $F8)+(background-160)            'top of girders with no ladder
            other:          return (_y & $F8)+7

PRI LadderBottomHeight(_x,_y) | background
    ' Calculates the position of the platform at the bottom of a ladder.  
    ' If we are above a ladder, then this one pixel above the platform at the bottom of the ladder,
    ' Otherwise it's the same value as LowPlatformHeight( ).
    ' If it's a broken ladder, then that doesn't count.  It's still the position of the platform above.
    repeat
        background := TileAt(_x,_y+8)
        debugdec(@tiletext,background)
        case background
            135..150,178:   _y+=8                                        'ladders + workaround invisible ladder
            152..158:       return (_y & $F8)+(background-152)           'platform with space below
            other:          return (_y & $F8)+7

PRI IsLadder(tile)
    if (tile=>135) and (tile=<150)
        return TRUE
    return FALSE
    
PRI IsEndOfLadder(tile)
    if (tile=>159) and (tile=<166)
        return TRUE
    return FALSE

PRI GroundHeight(x,y)
    
    
CON
' //////////////////////////////////////////////////////////////////
' NES Game Paddle Read
' //////////////////////////////////////////////////////////////////       
' reads both gamepads in parallel encodes 8-bits for each in format
' right game pad #1 [15..8] : left game pad #0 [7..0]

    GP_RIGHT  = %00000001
    GP_LEFT   = %00000010
    GP_DOWN   = %00000100
    GP_UP     = %00001000
    GP_START  = %00010000
    GP_SELECT = %00100000
    GP_B      = %01000000
    GP_A      = %10000000

PUB NES_Read_Gamepad : nes_bits        |  i

' set I/O ports to proper direction
' P3 = JOY_CLK      (4)
' P4 = JOY_SH/LDn   (5) 
' P5 = JOY_DATAOUT0 (6)
' P6 = JOY_DATAOUT1 (7)
' NES Bit Encoding

' step 1: set I/Os
DIRA [3] := 1 ' output
DIRA [4] := 1 ' output
DIRA [5] := 0 ' input
DIRA [6] := 0 ' input

' step 2: set clock and latch to 0
OUTA [3] := 0 ' JOY_CLK = 0
OUTA [4] := 0 ' JOY_SH/LDn = 0
'Delay(1)

' step 3: set latch to 1
OUTA [4] := 1 ' JOY_SH/LDn = 1
'Delay(1)

' step 4: set latch to 0
OUTA [4] := 0 ' JOY_SH/LDn = 0

' step 5: read first bit of each game pad

' data is now ready to shift out
' first bit is ready 
nes_bits := 0

' left controller
nes_bits := INA[5] | (INA[6] << 8)

' step 7: read next 7 bits
repeat i from 0 to 6
 OUTA [3] := 1 ' JOY_CLK = 1
 'Delay(1)             
 OUTA [3] := 0 ' JOY_CLK = 0
 nes_bits := (nes_bits << 1)
 nes_bits := nes_bits | INA[5] | (INA[6] << 8)

 'Delay(1)             
' invert bits to make positive logic
nes_bits := (!nes_bits & $FFFF)

' //////////////////////////////////////////////////////////////////
' End NES Game Paddle Read
' //////////////////////////////////////////////////////////////////  

VAR
  long  row
  long  col

DAT

laddertoptext           byte    "LADDERTOP",0
ladderbottomtext        byte    "LADDERBOTTOM",0
tiletext                byte    "TILE",0
platformheighttext      byte    "PLATFORMHEIGHT",0
deltatext               byte    "DELTATEXT",0
currentheighttext       byte    "CURRENTHEIGHT",0
ypostext                byte    "YPOS",0
errortext               byte    "ERROR",0
kongstatetext           byte    "KONGSTATE",0
kongstatecounttext      byte    "KONGSTATECOUNT",0
barrelstatetext         byte    "BARRELSTATE",0
kongbarreltext          byte    "KONGHOLDINGBARREL",0
heighttofalltext        byte    "HEIGHTTOFALL",0
knongbarrelreleasestext byte    "KONGBARRELRELEASES",0

PUB DebugInit(map)
    bytefill(map,151,32*28)
    DebugReset

PUB DebugReset
  row := 4
  col := 0

PUB DebugHex(title,number) | adr, i
    i:=0
    repeat while byte[title][i]<>0
        debug_map[row*32+col++] := byte[title][i]-65+10
        i++
    debug_map[row*32+col++] := 39
    debug_map[row*32+col++] := number >> 28   
    debug_map[row*32+col++] := (number >> 24) & $F
    debug_map[row*32+col++] := (number >> 20) & $F
    debug_map[row*32+col++] := (number >> 16) & $F
    debug_map[row*32+col++] := (number >> 12) & $F
    debug_map[row*32+col++] := (number >> 8) & $F
    debug_map[row*32+col++] := (number >> 4) & $F
    debug_map[row*32+col++] := number & $F
    col:=0
    row+=1
    row<#=27

PUB DebugDec(title,number) | adr,out[10], i
    i:=0
    repeat while byte[title][i]<>0
        debug_map[row*32+col++] := byte[title][i]-65+10
        i++
    debug_map[row*32+col++] := 39
    debug_map[row*32+col+9] := number//10 + 99
    number/=10   
    debug_map[row*32+col+8] := number//10 + 99
    number/=10   
    debug_map[row*32+col+7] := number//10 + 99
    number/=10   
    debug_map[row*32+col+6] := number//10 + 99
    number/=10   
    debug_map[row*32+col+5] := number//10 + 99
    number/=10   
    debug_map[row*32+col+4] := number//10 + 99
    number/=10   
    debug_map[row*32+col+3] := number//10 + 99
    number/=10            
    debug_map[row*32+col+2] := number//10 + 99
    number/=10   
    debug_map[row*32+col+1] := number//10 + 99
    number/=10   
    debug_map[row*32+col+0] := number//10 + 99
    col:=0
    row+=1
    row<#=27
  
DAT

scanline_req            long    0

tv_params               long    @scanline_req+$10            'status
                        long    @hub_scanline_buffer+$10     'enable

' Hub scanline buffer = 256 pixels, 8 bits per pixel = 2048 bits = 256 bytes = 64 longs
' However, sprites are going to be 16x16 and I want the facility to be able to slide sprites off
' screen to the left or right without explicit clipping code or corrupting adjacent cog_memory
' So an extra 16 pixels = 128 bits = 4 longs either side of the visible scan buffer
                        long     0
                        long     0
                        long     0
                        long     0
                                     
hub_scanline_buffer
colorbar0               long    $0b_0b_0b_0b
colorbar1               long    $0c_0c_0c_0c
colorbar2               long    $0d_0d_0d_0d
colorbar3               long    $0e_0e_0e_0e
colorbar4               long    $1b_0b_1b_1b
colorbar5               long    $1c_0c_1c_1c
colorbar6               long    $1d_0d_1d_1d
colorbar7               long    $1e_0e_1e_1e
colorbar8               long    $2b_0b_2b_2b
colorbar9               long    $2c_2c_2c_2c
colorbar10              long    $2d_2d_2d_2d
colorbar11              long    $2e_2e_2e_2e
colorbar12              long    $3b_3b_3b_3b
colorbar13              long    $3c_3c_3c_3c
colorbar14              long    $3d_3d_3d_3d
colorbar15              long    $3e_3e_3e_3e
colorbar16              long    $4b_4b_4b_4b
colorbar17              long    $4c_4c_4c_4c
colorbar18              long    $4d_4d_4d_4d
colorbar19              long    $4e_4e_4e_4e
colorbar20              long    $5b_5b_5b_5b
colorbar21              long    $5c_5c_5c_5c
colorbar22              long    $5d_5d_5d_5d
colorbar23              long    $5e_5e_5e_5e
colorbar24              long    $6b_6b_6b_6b
colorbar25              long    $6c_6c_6c_6c
colorbar26              long    $6d_6d_6d_6d
colorbar27              long    $6e_6e_6e_6e
colorbar28              long    $7b_7b_7b_7b
colorbar29              long    $7c_7c_7c_7c
colorbar30              long    $7d_7d_7d_7d
colorbar31              long    $7e_7e_7e_7e
colorbar32              long    $8b_8b_8b_8b
colorbar33              long    $8c_8c_8c_8c
colorbar34              long    $8d_8d_8d_8d
colorbar35              long    $8e_8e_8e_8e
colorbar36              long    $9b_9b_9b_9b
colorbar37              long    $9c_9c_9c_9c
colorbar38              long    $9d_9d_9d_9d
colorbar39              long    $9e_9e_9e_9e
colorbar40              long    $ab_ab_ab_ab
colorbar41              long    $ac_ac_ac_ac
colorbar42              long    $ad_ad_ad_ad
colorbar43              long    $ae_ae_ae_ae
colorbar44              long    $bb_bb_bb_bb
colorbar45              long    $bc_bc_bc_bc
colorbar46              long    $bd_bd_bd_bd
colorbar47              long    $be_be_be_be
colorbar48              long    $cb_cb_cb_cb
colorbar49              long    $cc_cc_cc_cc
colorbar50              long    $cd_cd_cd_cd
colorbar51              long    $ce_ce_ce_ce
colorbar52              long    $db_db_db_db
colorbar53              long    $dc_dc_dc_dc
colorbar54              long    $dd_dd_dd_dd
colorbar55              long    $de_de_de_de
colorbar56              long    $eb_eb_eb_eb
colorbar57              long    $ec_ec_ec_ec
colorbar58              long    $ed_ed_ed_ed
colorbar59              long    $ee_ee_ee_ee
colorbar60              long    $fb_fb_fb_fb
colorbar61              long    $fc_fc_fc_fc
colorbar62              long    $fd_fd_fd_fd
colorbar63              long    $fe_fe_fe_fe

                        long    0
                        long    0
                        long    0
                        long    0
' End of scanline buffer